# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds the ExecutionPluginManager as well as the base-class
for all ExecutionPlugins.
"""

from .pluginmanager import PluginSubManager
from ...core.interface import Interface


class InterfacePluginManager(PluginSubManager):
    """
    Container holding all the CollectorPlugins
    """

    def __init__(self, parent):
        """
        Create the Coll
        :param path:
        :param recursive:
        :return:
        """
        super(InterfacePluginManager, self).__init__(parent=parent,
                                                     plugin_class=Interface)

    @property
    def _instantiate(self):
        """
        Indicate that the plugins should not instantiated before being stored
        """
        return False
